#define _XOPEN_SOURCE 500

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <unistd.h>
#include <sys/syslog.h>
#include <time.h>
#include <errno.h>
#include <sqlite3.h>
#include <uuid/uuid.h>
#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>

#ifdef WITH_SOCKET_ACTIVATION

#include <systemd/sd-daemon.h>

#endif

#include "bytemess.h"
#include "shared.h"
#include "parser.h"

char *dbname = ":memory:";
char *pidfile = NULL;
int reduced_writes = 0;
int garbage_collect_threshold = -1;
int signal_pipe[2];
sqlite3 *db;
sqlite3_stmt *insert_query;
long seq = 0;
long last_kseq = 0;

char *logsocket = "/dev/log";
char *logreadsocket = LOGREAD_SOCKET;
int logread_fd;
#define LOGREAD_MAX_CLIENTS 10
int logread_client[LOGREAD_MAX_CLIENTS];

#define SCHEMA_VERSION 1

uuid_t boot_id;
int verbose = 0;

void
createPidfile(char *pidfile);

void
usage(char *name)
{
	// vfd:l:
	fprintf(stderr, "Usage: %s [-v] [-f] [-r] [-d path] [-l path] [-g kb]\n", name);
	fprintf(stderr, "The sqlite-backed syslog daemon\n\n");
	fprintf(stderr, "Arguments:\n");
	fprintf(stderr, "  -v          Show verbose output\n");
	fprintf(stderr, "  -f          Run the process in the foreground\n");
	fprintf(stderr, "  -d path     Path to the sqlite database or :memory:\n");
	fprintf(stderr, "  -l path     Location for the syslog socket (/dev/log)\n");
	fprintf(stderr, "  -r          Reduced write mode, only save database on exit\n");
	fprintf(stderr, "  -g kB       Remove old log lines when the in-memory database crosses x kB\n");
	fprintf(stderr, "  -h          Display this help text\n");
}

void
verbosef(const char *fmt, ...)
{
	if (verbose == 0) return;
	va_list args;
	va_start(args, fmt);
	vprintf(fmt, args);
	va_end(args);
}

int
pragma(char *query)
{
	char *err = 0;

	int ret = sqlite3_exec(db, query, NULL, NULL, &err);
	if (ret != SQLITE_OK) {
		printf("%s: %s\n", sqlite3_errstr(sqlite3_extended_errcode(db)), sqlite3_errmsg(db));
		return -1;
	}
	return 0;
}

void
prepare_insert()
{
	char *sql = "INSERT INTO log (message, sent, received, severity, facility, host, app, procid, msgid, fields, boot_id, seq, kseq) VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?);";
	sqlite3_prepare_v2(db, sql, -1, &insert_query, NULL);
}

int
garbage_collect()
{
	int cur = -1;
	int unused;
	sqlite3_db_status(db, SQLITE_DBSTATUS_CACHE_USED, &cur, &unused, 0);
	if (cur > garbage_collect_threshold) {
		sqlite3_finalize(insert_query);

		char *sql = "DELETE FROM log WHERE ROWID IN (SELECT ROWID FROM log WHERE facility=? ORDER BY ROWID DESC LIMIT -1 OFFSET 1000)";
		sqlite3_stmt *stmt;
		sqlite3_prepare_v2(db, sql, -1, &stmt, NULL);
		for (int facility = 0; facility < 22; facility++) {
			sqlite3_bind_int(stmt, 1, facility);
			if (sqlite3_step(stmt) != SQLITE_DONE) {
				printf("%s: %s\n", sqlite3_errstr(sqlite3_extended_errcode(db)), sqlite3_errmsg(db));
			}
			sqlite3_clear_bindings(stmt);
			sqlite3_reset(stmt);
		}
		sqlite3_finalize(stmt);
		pragma("VACUUM;");
		prepare_insert();
	}
	return 0;
}

int
log_insert(int src, char *msg, long sent, int sev, int fac, char *hostname, char *app, char *procid,
	char *msgid, char *fields)
{
	time_t received;
	time(&received);
	if (verbose) {
		if (src == SRC_UNKNOWN) {
			printf("[unknown] %s\n", msg);
		} else if (src == SRC_RFC5424) {
			printf("[RFC5424 %s %s]", severity_name[sev], facility_name[fac]);
			if (hostname[0] != '-') {
				printf(" %s", hostname);
			}
			if (app[0] != '-') {
				printf(" %s", app);
			}
			if (procid[0] != '-') {
				printf(" %s", procid);
			}
			if (msgid[0] != '-') {
				printf(" %s", msgid);
			}
			printf(": %s\n", msg);
		} else if (src == SRC_RFC3164) {
			printf("[RFC3164 %s %s]", severity_name[sev], facility_name[fac]);
			if (app[0] != '-') {
				printf(" %s", app);
			}
			printf(": %s\n", msg);
		} else if (src == SRC_KMSG) {
			printf("[KMSG %s %s]", severity_name[sev], facility_name[fac]);
			if (app[0] != '-') {
				printf(" %s", app);
			}
			printf(": %s\n", msg);
		}
	}

	unsigned char packet[MAX_MESSAGE_SIZE];
	unsigned char *packet_ptr = packet;
	int msglen = (int) strlen(msg);
	packet_ptr = write_int8(packet_ptr, PACKET_LIVE);
	packet_ptr = write_int64(packet_ptr, received);
	packet_ptr = write_int8(packet_ptr, sev);
	packet_ptr = write_int8(packet_ptr, fac);
	packet_ptr = write_varchar(packet_ptr, app);
	memcpy(packet_ptr, msg, msglen);
	packet_ptr += msglen;
	*packet_ptr++ = '\0';
	for (int i = 0; i < LOGREAD_MAX_CLIENTS; i++) {
		if (logread_client[i] == 0)
			continue;
		ssize_t written = write(logread_client[i], packet, packet_ptr - packet);
		if (written < 0) {
			perror("write");
		}
	}

	sqlite3_bind_text(insert_query, 1, msg, (int) strlen(msg), NULL);
	if (sent != -1) {
		sqlite3_bind_int64(insert_query, 2, sent);
	} else {
		sqlite3_bind_null(insert_query, 2);
	}
	sqlite3_bind_int64(insert_query, 3, received);
	sqlite3_bind_int(insert_query, 4, sev);
	sqlite3_bind_int(insert_query, 5, fac);
	if (hostname == NULL || hostname[0] == '-') {
		sqlite3_bind_null(insert_query, 6);
	} else {
		sqlite3_bind_text(insert_query, 6, hostname, (int) strlen(hostname), NULL);
	}
	if (app == NULL || app[0] == '-') {
		sqlite3_bind_null(insert_query, 7);
	} else {
		sqlite3_bind_text(insert_query, 7, app, (int) strlen(app), NULL);
	}
	if (procid == NULL || procid[0] == '-') {
		sqlite3_bind_null(insert_query, 8);
	} else {
		sqlite3_bind_text(insert_query, 8, procid, (int) strlen(procid), NULL);
	}
	if (msgid == NULL || msgid[0] == '-') {
		sqlite3_bind_null(insert_query, 9);
	} else {
		sqlite3_bind_text(insert_query, 9, msgid, (int) strlen(msgid), NULL);
	}
	if (fields == NULL || fields[0] == '-') {
		sqlite3_bind_null(insert_query, 10);
	} else {
		sqlite3_bind_text(insert_query, 10, fields, (int) strlen(fields), NULL);
	}
	sqlite3_bind_blob(insert_query, 11, boot_id, 16, SQLITE_STATIC);
	sqlite3_bind_int64(insert_query, 12, seq++);
	if (src == SRC_KMSG) {
		sqlite3_bind_int64(insert_query, 13, last_kseq);
	} else {
		sqlite3_bind_null(insert_query, 13);
	}
	if (sqlite3_step(insert_query) != SQLITE_DONE) {
		printf("%s: %s\n", sqlite3_errstr(sqlite3_extended_errcode(db)), sqlite3_errmsg(db));
	}
	sqlite3_clear_bindings(insert_query);
	sqlite3_reset(insert_query);
	if (garbage_collect_threshold > 0) {
		garbage_collect();
	}
	return 0;
}

void
self_log(int sev, const char *fmt, ...)
{
	char buffer[512];
	va_list args;
	va_start(args, fmt);
	vsprintf(buffer, fmt, args);
	va_end(args);
	if (sev < 6) {
		fprintf(stderr, "%s\n", buffer);
	} else {
		verbosef("%s\n", buffer);
	}
	log_insert(SRC_SELF, buffer, -1, sev, 5, "-", "logbookd", "-", "-", "-");
}

int
create_dev_log()
{
	struct sockaddr_un addr;
	int fd;
	memset(&addr, 0, sizeof(addr));
	addr.sun_family = AF_UNIX;
	strcpy(addr.sun_path, logsocket);

	unlink(addr.sun_path);
	verbosef("Creating log socket at %s\n", addr.sun_path);
	fd = socket(AF_UNIX, SOCK_DGRAM, 0);
	if (fd < 0) {
		fprintf(stderr, "Could not create log socket: [%d] %s\n", errno, strerror(errno));
		return fd;
	}
	int ret = bind(fd, (struct sockaddr *) &addr, strlen(addr.sun_path) + sizeof(addr.sun_family));
	if (ret < 0) {
		fprintf(stderr, "Could not bind log socket: [%d] %s\n", errno, strerror(errno));
		return ret;
	}
	if (chmod(addr.sun_path, 0666) != 0) {
		fprintf(stderr, "Couldn't chmod %s: [%d] %s\n", addr.sun_path, errno, strerror(errno));
		return -1;
	}
	return fd;
}

int
create_logread_fd()
{
	struct sockaddr_un addr;
	int fd;
	memset(&addr, 0, sizeof(addr));
	addr.sun_family = AF_UNIX;
	strcpy(addr.sun_path, logreadsocket);

	unlink(addr.sun_path);
	verbosef("Creating log-read socket at %s\n", addr.sun_path);
	fd = socket(AF_UNIX, SOCK_SEQPACKET, 0);
	if (fd < 0) {
		fprintf(stderr, "Could not create log-read socket: [%d] %s\n", errno, strerror(errno));
		return fd;
	}
	int ret = bind(fd, (struct sockaddr *) &addr, strlen(addr.sun_path) + sizeof(addr.sun_family));
	if (ret < 0) {
		fprintf(stderr, "Could not bind log-read socket: [%d] %s\n", errno, strerror(errno));
		return ret;
	}
	if (chmod(addr.sun_path, 0666) != 0) {
		fprintf(stderr, "Couldn't chmod %s: [%d] %s\n", addr.sun_path, errno, strerror(errno));
		return -1;
	}
	return fd;
}

void
parse_kmsg_line(char *buffer, ssize_t size)
{
	buffer[size] = '\0';
	char *end = NULL;
	long pri = strtol(buffer, &end, 10);
	if (end && end > buffer) {
		buffer = end + 1;
	}
	end = NULL;
	long kseq = strtol(buffer, &end, 10);
	if (end && end > buffer) {
		buffer = end + 1;
	}

	// Don't log kernel messages that already have been logged this boot
	if (kseq <= last_kseq) {
		return;
	}
	last_kseq = kseq;

	end = NULL;
	long ts = strtol(buffer, &end, 10);
	if (end && end > buffer) {
		buffer = end + 1;
	}

	long flags = 0;
	if (buffer[0] == '-') {
		buffer += 1;
	} else {
		end = NULL;
		flags = strtol(buffer, &end, 10);
		if (end && end > buffer) {
			buffer = end;
		}
	}

	while (*buffer != ';') {
		buffer++;
	}
	buffer++;

	int i = 0;
	while (buffer[i++] != '\n') {
	}
	i--;
	buffer[i] = '\0';
	char *fields = &buffer[i + 1];

	int fac = (int) pri >> 3;
	int sev = (int) pri & 7;
	log_insert(SRC_KMSG, buffer, ts, sev, fac, NULL, "kernel", NULL, NULL, fields);
}

void
parse_syslog_pri_and_log(char *buffer, ssize_t size)
{
	buffer[size] = '\0';
	char *p = buffer;
	buffer += size;

	while (p < buffer) {
		// The parsed and escaped buffer can be twice the input buffer size due to
		// escaping certain characters to two characters
		char parsebuf[MAX_MESSAGE_SIZE * 2];
		char *res = parsebuf;
		char c;

		int pri = (LOG_USER | LOG_NOTICE);
		if (*p == '<') {
			pri = (int) strtol(p + 1, &p, 10);
			if (*p == '>')
				p++;
			if (pri & ~(LOG_FACMASK | LOG_PRIMASK))
				pri = (LOG_USER | LOG_NOTICE);
		}

		while ((c = *p++)) {
			if (c == '\n')
				c = ' ';

			// Convert special characters into escapes
			if ((!(c & ~0x1f) && c != '\t')) {
				*res++ = '^';
				c += '@';
			}
			*res++ = c;
		}

		*res = '\0';

		struct logmessage parsed = {0};
		parse_log_message(pri, parsebuf, res - parsebuf, &parsed);
		log_insert(parsed.format, parsed.msg, parsed.sent, parsed.sev, parsed.fac, parsed.hostname, parsed.app,
			parsed.procid, parsed.msgid, parsed.fields);
	}
}

void
do_query(int client, unsigned char *cmd, ssize_t size)
{
	cmd[size] = '\0';
	int limit;
	char unit[256] = "";
	cmd = read_int32(cmd, &limit);
	cmd = read_varchar(cmd, unit);

	char sql[1024] = "SELECT * FROM ( SELECT seq, received, severity, facility, app, message FROM log ";
	if (strlen(unit)) {
		strcat(sql, "WHERE app LIKE $app || '%' ");
	}
	strcat(sql, "ORDER BY seq DESC LIMIT $limit) ORDER BY seq ASC");
	sqlite3_stmt *stmt;

	int ret = sqlite3_prepare_v2(db, sql, -1, &stmt, NULL);
	if (ret != SQLITE_OK) {
		printf("%s: %s\n", sqlite3_errstr(sqlite3_extended_errcode(db)), sqlite3_errmsg(db));
	}
	if (strlen(unit)) {
		sqlite3_bind_text(stmt, sqlite3_bind_parameter_index(stmt, "$app"), unit, (int) strlen(unit), NULL);
	}
	sqlite3_bind_int(stmt, sqlite3_bind_parameter_index(stmt, "$limit"), limit);
	unsigned char packet[MAX_MESSAGE_SIZE];
	while (sqlite3_step(stmt) == SQLITE_ROW) {
		unsigned char *packet_ptr = packet;
		packet_ptr = write_int8(packet_ptr, PACKET_ROW);
		packet_ptr = write_int64(packet_ptr, sqlite3_column_int64(stmt, 1));
		packet_ptr = write_int8(packet_ptr, sqlite3_column_int(stmt, 2));
		packet_ptr = write_int8(packet_ptr, sqlite3_column_int(stmt, 3));
		const unsigned char *app = (unsigned char *) sqlite3_column_text(stmt, 4);
		const unsigned char *msg = sqlite3_column_text(stmt, 5);
		int msglen = (int) strlen((char *) msg);
		packet_ptr = write_varchar(packet_ptr, (char *) app);
		memcpy(packet_ptr, msg, msglen);
		packet_ptr += msglen;
		*packet_ptr++ = '\0';
		write(client, packet, packet_ptr - packet);
	}
	unsigned char *packet_ptr = packet;
	packet_ptr = write_int8(packet_ptr, PACKET_DONE);
	sqlite3_finalize(stmt);
	write(client, packet, packet_ptr - packet);
}

static int
query_result(void *NotUsed, int argc, char **argv, char **azColName)
{
	// Callback function for sqlite3 to display the query result from
	// the sqlite3_exec() call
	int i;
	for (i = 0; i < argc; i++) {
		printf("%s = %s\n", azColName[i], argv[i] ? argv[i] : "NULL");
	}
	printf("\n");
	return 0;
}

int
get_user_version()
{
	char *sql = "PRAGMA user_version;";
	sqlite3_stmt *stmt;
	int result = 0;

	sqlite3_prepare_v2(db, sql, -1, &stmt, NULL);
	sqlite3_step(stmt);
	result = sqlite3_column_int(stmt, 0);
	sqlite3_finalize(stmt);
	return result;
}

long
get_last_seq()
{
	char *sql = "SELECT seq FROM log ORDER BY seq DESC LIMIT 1;";
	sqlite3_stmt *stmt;
	long result = 0;

	sqlite3_prepare_v2(db, sql, -1, &stmt, NULL);
	int ret = sqlite3_step(stmt);
	if (ret == SQLITE_ROW) {
		result = sqlite3_column_int64(stmt, 0);
	}
	sqlite3_finalize(stmt);
	return result;
}

long
get_last_kseq()
{
	char *sql = "SELECT kseq FROM log WHERE boot_id=? ORDER BY kseq DESC LIMIT 1;";
	sqlite3_stmt *stmt;
	long result = 0;

	sqlite3_prepare_v2(db, sql, -1, &stmt, NULL);
	sqlite3_bind_blob(stmt, 1, boot_id, 16, SQLITE_STATIC);
	if (sqlite3_step(stmt) == SQLITE_ROW) {
		result = sqlite3_column_int64(stmt, 0);
	}
	sqlite3_finalize(stmt);
	return result;
}

int
set_user_version(int version)
{
	char query[32];
	snprintf(query, 32, "PRAGMA user_version = %d;", version);
	return pragma(query);
}

int
make_schema_v1()
{
	char *err = 0;
	char *sql = "CREATE TABLE log(" \
                "   seq INT NOT NULL," \
                "   kseq INT," \
                "	message TEXT NOT NULL," \
                "	boot_id BLOB NOT NULL," \
                "	host TEXT," \
                "	app TEXT," \
                "	pid INT," \
                "	procid TEXT," \
                "	msgid TEXT," \
                "	fields TEXT," \
                "	sent INT," \
                "	received INT NOT NULL," \
                "	severity INT NOT NULL," \
                "	facility INT NOT NULL);";
	int ret = sqlite3_exec(db, sql, query_result, 0, &err);
	if (ret != SQLITE_OK) {
		printf("ERR[%d]: %s\n", ret, err);
		return -1;
	}
	return set_user_version(1);
}

void
init_boot_id()
{
	FILE *fp = fopen("/proc/sys/kernel/random/boot_id", "r");
	char raw[37];
	fgets(raw, 37, fp);
	raw[36] = '\0';
	verbosef("boot_id is %s\n", raw);
	if (uuid_parse(raw, boot_id) != 0) {
		printf("Could not read boot_id\n");
	}
}

int
restore()
{
	if (access(dbname, F_OK) != 0) {
		fprintf(stderr, "Cannot restore %s\n", dbname);
		return 0;
	}
	sqlite3 *backup_db;
	if (sqlite3_open(dbname, &backup_db)) {
		fprintf(stderr, "Can't open database: %s\n", sqlite3_errmsg(backup_db));
		return -1;
	}
	sqlite3_backup *bak = sqlite3_backup_init(db, "main", backup_db, "main");
	if (!bak) {
		fprintf(stderr, "Could not start data restore: %s\n", sqlite3_errmsg(db));
		return -1;
	}

	sqlite3_backup_step(bak, -1);
	sqlite3_backup_finish(bak);
	sqlite3_close(backup_db);
	return 0;
}

int
backup()
{
	sqlite3 *backup_db;
	if (sqlite3_open(dbname, &backup_db)) {
		fprintf(stderr, "Can't open database: %s\n", sqlite3_errmsg(backup_db));
		return -1;
	}

	sqlite3_finalize(insert_query);

	sqlite3_backup *bak = sqlite3_backup_init(backup_db, "main", db, "main");
	if (!bak) {
		fprintf(stderr, "Could not start data backup: %s\n", sqlite3_errmsg(db));
		return -1;
	}

	sqlite3_backup_step(bak, -1);
	sqlite3_backup_finish(bak);
	sqlite3_close(backup_db);

	prepare_insert();
	return 0;
}

void
signal_to_pipe(int sig)
{
	int old_errno = errno;
	char buf[1] = {(char) sig};
	if (write(signal_pipe[1], buf, 1) == -1) {
		if (sig == SIGTERM || sig == SIGINT) {
			exit(2);
		}
	}
	errno = old_errno;
}

int
make_nonblock(int fd)
{
	int flags = fcntl(fd, F_GETFL);
	if (flags == -1)
		return flags;

	flags |= O_NONBLOCK;
	if (fcntl(fd, F_SETFL, flags) == -1)
		return -1;
	return 0;
}

int
mkpath(char *path, mode_t mode)
{
	for (char *p = strchr(path + 1, '/'); p; p = strchr(p + 1, '/')) {
		*p = '\0';
		if (mkdir(path, mode) == -1) {
			if (errno != EEXIST) {
				*p = '/';
				return -1;
			}
		}
		*p = '/';
	}
	return 0;
}

void
createPidfile(char *path)
{
	int fd;
	fd = open(path, O_RDWR | O_CREAT, 0644);
	if (fd == -1) {
		fprintf(stderr, "Could not create pid file %s: %s\n", path, strerror(errno));
		exit(1);
	}
	FILE *f = fdopen(fd, "r+");
	int pid = getpid();
	if (!fprintf(f, "%d\n", pid)) {
		fprintf(stderr, "write not create pid file %s: %s\n", path, strerror(errno));
		exit(1);
	}
	fflush(f);
}


int
main(int argc, char *argv[])
{
	int c;
	int daemonize = 1;
	opterr = 0;
	char *end;
	long res;

	while ((c = getopt(argc, argv, "vfhrg:d:l:p:")) != -1) {
		switch (c) {
			case 'v':
				verbose = 1;
				break;
			case 'd':
				dbname = optarg;
				break;
			case 'l':
				logsocket = optarg;
				break;
			case 'p':
				pidfile = optarg;
				break;
			case 'f':
				daemonize = 0;
				break;
			case 'h':
				usage(argv[0]);
				return 0;
			case 'r':
				reduced_writes = 1;
				break;
			case 'g':
				res = strtol(optarg, &end, 10);
				if (end == optarg || end == NULL || *end != '\0') {
					fprintf(stderr, "Invalid value for -g, number required\n");
					exit(1);
				}
				garbage_collect_threshold = ((int) res) * 1024;
			case '?':
				if (optopt == 'd' || optopt == 'l') {
					fprintf(stderr, "Option -%c requires an argument.\n", optopt);
				} else if (isprint(optopt)) {
					fprintf(stderr, "Unknown option '-%c'\n", optopt);
				} else {
					fprintf(stderr, "Unknown option character x%x\n", optopt);
				}
				return 1;
			default:
				usage(argv[0]);
				return 1;
		}
	}

	init_boot_id();
	verbosef("Using database %s\n", dbname);
	if (reduced_writes) {
		if (sqlite3_open(":memory:", &db)) {
			printf("Can't open database: %s\n", sqlite3_errmsg(db));
			return 1;
		}
		verbosef("Running in reduced writes mode, restoring old data...");
		if (restore() != 0) {
			return 1;
		}
	} else {
		if (sqlite3_open(dbname, &db)) {
			printf("Can't open database: %s\n", sqlite3_errmsg(db));
			return 1;
		}
	}
	int ver = get_user_version();
	verbosef("Current schema version is %d\n", ver);
	if (ver != SCHEMA_VERSION) {
		make_schema_v1();
	}

	prepare_insert();

	char *tmp;
	tmp = strdup(logreadsocket);
	if (mkpath(tmp, 0755) != 0) {
		fprintf(stderr, "Could not create directory for log-read socket\n");
		exit(1);
	}

	int log_fd;
#ifdef WITH_SOCKET_ACTIVATION
	int n_sockets = sd_listen_fds(0);
	if (n_sockets > 1) {
		self_log(SEV_CRIT, "Too many file descriptors received");
		return 1;
	} else if (n_sockets == 1) {
		self_log(SEV_INFO, "Started by socket activation");
		log_fd = SD_LISTEN_FDS_START + 0;
		if (!sd_is_socket_unix(log_fd, SOCK_DGRAM, -1, NULL, 0)) {
			self_log(SEV_CRIT, "fd received from socket activation is not a unix socket");
		}
	} else {
#endif
		log_fd = create_dev_log();
		if (log_fd < 0) {
			return 1;
		}
		self_log(SEV_INFO, "Started syslog daemon at %s", logsocket);
#ifdef WITH_SOCKET_ACTIVATION
	}
#endif
	logread_fd = create_logread_fd();
	if (logread_fd < 0) {
		return 1;
	}
	if (listen(logread_fd, 20) == -1) {
		fprintf(stderr, "Could not listen: [%d] %s", errno, strerror(errno));
	}
	int kmsg_fd = open("/dev/kmsg", O_RDONLY);
	if (kmsg_fd < 0) {
		self_log(SEV_ERR, "Could not open kernel log: [%d] %s", errno, strerror(errno));
	} else {
		self_log(SEV_INFO, "Opened kernel log at /dev/kmsg");
	}

	self_log(SEV_INFO, "Storing logs in %s", dbname);

	if (daemonize > 0) {
		// First fork
		pid_t pid = fork();
		if (pid < 0) {
			fprintf(stderr, "First fork failed\n");
			return 1;
		}
		if (pid > 0) {
			// This is the parent process
			return 0;
		}
		if (setsid() < 0) {
			return 1;
		}

		// Second fork
		pid = fork();
		if (pid < 0) {
			return 1;
		}
		if (pid > 0) {
			// This is the parent process
			return 0;
		}

		umask(0);
		chdir("/");
		close(1);
		close(2);
	}

	if (pidfile != NULL) {
		createPidfile(pidfile);
	}

	// Make a pipe to send signals to the select() loop
	if (pipe(signal_pipe) == -1) {
		fprintf(stderr, "Could not create signal pipe\n");
		return 1;
	}
	make_nonblock(signal_pipe[0]);
	make_nonblock(signal_pipe[1]);

	struct sigaction act;
	act.sa_handler = signal_to_pipe;
	act.sa_flags = SA_NOCLDSTOP;
	sigaction(SIGINT, &act, NULL);
	sigaction(SIGTERM, &act, NULL);
	sigaction(SIGUSR1, &act, NULL);

	seq = get_last_seq();
	last_kseq = get_last_kseq();
	self_log(SEV_DEBUG, "Ready to process log messages");

	// The main loop
	char recvbuf[MAX_MESSAGE_SIZE];
	for (int i = 0; i < LOGREAD_MAX_CLIENTS; i++) {
		logread_client[i] = 0;
	}
	while (1) {
		fd_set read_fd_set;
		fd_set err_fd_set;
		FD_ZERO(&read_fd_set);
		FD_ZERO(&err_fd_set);
		FD_SET(signal_pipe[0], &read_fd_set);
		FD_SET(log_fd, &read_fd_set);
		FD_SET(log_fd, &err_fd_set);
		FD_SET(logread_fd, &read_fd_set);
		for (int i = 0; i < LOGREAD_MAX_CLIENTS; i++) {
			if (logread_client[i] != 0) {
				FD_SET(logread_client[i], &read_fd_set);
			}
		}
		if (kmsg_fd > 0) {
			FD_SET(kmsg_fd, &read_fd_set);
		}
		int ret = select(FD_SETSIZE, &read_fd_set, NULL, &err_fd_set, NULL);
		if (ret < 0) {
			if (errno == EINTR) {
				continue;
			}
			self_log(SEV_CRIT, "select() fail");
			break;
		}

		if (FD_ISSET(signal_pipe[0], &read_fd_set)) {
			char sig[2] = "\0";
			while (1) {
				if (read(signal_pipe[0], &sig, 1) == -1) {
					break;
				}
				if (sig[0] == SIGINT || sig[0] == SIGTERM) {
					fprintf(stderr, "Received interrupt, flushing database\n");
					sqlite3_finalize(insert_query);

					if (reduced_writes) {
						backup();
					}

					sqlite3_close(db);
					exit(0);
				} else if (sig[0] == SIGUSR1) {
					if (reduced_writes) {
						self_log(SEV_INFO, "Received USR1. Writing out database");
						backup();
					} else {
						self_log(SEV_ERR, "Received USR1 but not running in reduced-write mode");
					}
				} else {
					self_log(SEV_CRIT, "Received unhandled signal");
				}
			}
		}

		if (FD_ISSET(log_fd, &err_fd_set)) {
			self_log(SEV_CRIT, "Exception on syslog fd");
		}

		if (FD_ISSET(log_fd, &read_fd_set)) {
			ssize_t size;
			size = read(log_fd, recvbuf, MAX_MESSAGE_SIZE - 1);
			parse_syslog_pri_and_log(recvbuf, size);
		}
		if (kmsg_fd > 0 && FD_ISSET(kmsg_fd, &read_fd_set)) {
			ssize_t size;
			size = read(kmsg_fd, recvbuf, MAX_MESSAGE_SIZE - 1);
			parse_kmsg_line(recvbuf, size);
		}
		if (FD_ISSET(logread_fd, &read_fd_set)) {
			int new_client_fd = accept(logread_fd, NULL, NULL);
			if (new_client_fd < 0) {
				self_log(SEV_ERR, "Could not accept client: [%d] %s", errno, strerror(errno));
			}
			int full = 1;
			for (int i = 0; i < LOGREAD_MAX_CLIENTS; i++) {
				if (logread_client[i] == 0) {
					logread_client[i] = new_client_fd;
					full = 0;
					break;
				}
			}
			if (full) {
				close(new_client_fd);
			}
		}
		for (int i = 0; i < LOGREAD_MAX_CLIENTS; i++) {
			if (logread_client[i] == 0)
				continue;

			if (FD_ISSET(logread_client[i], &read_fd_set)) {
				ssize_t size;
				size = read(logread_client[i], recvbuf, MAX_MESSAGE_SIZE - 1);
				if (size < 1) {
					close(logread_client[i]);
					logread_client[i] = 0;
					continue;
				}
				if (size < 0) {
					self_log(SEV_ERR, "Could not read from client: [%d] %s", errno, strerror(errno));
				}
				do_query(logread_client[i], (unsigned char *) recvbuf, size);
			}
		}
	}
	return 0;
}
